@@ -1,4 +1,5 @@
 from __future__ import absolute_import, print_function, division
+from pony.py23compat import PY38
 
 import ast
 
@@ -106,19 +107,26 @@ class PythonTranslator(ASTTranslator):
     def postExpr(translator, node):
         return node.value.src
     def postIfExp(translator, node):
-        return '%s if %s else %s' % (node.then.src, node.test.src, node.else_.src)
+        return '%s if %s else %s' % (node.body.src, node.test.src, node.orelse.src)
     def postLambda(translator, node):
-        argnames = list(node.argnames)
-        kwargs_name = argnames.pop() if node.kwargs else None
-        varargs_name = argnames.pop() if node.varargs else None
-        def_argnames = argnames[-len(node.defaults):] if node.defaults else []
-        nodef_argnames = argnames[:-len(node.defaults)] if node.defaults else argnames
-        args = ', '.join(nodef_argnames)
-        d_args = ', '.join('%s=%s' % (argname, default.src) for argname, default in zip(def_argnames, node.defaults))
-        v_arg = '*%s' % varargs_name if varargs_name else None
-        kw_arg = '**%s' % kwargs_name if kwargs_name else None
-        args = ', '.join(x for x in [args, d_args, v_arg, kw_arg] if x)
-        return 'lambda %s: %s' % (args, node.code.src)
+        return 'lambda %s: %s' % (node.args.src, node.body.src)
+    def postarguments(translator, node):
+        if node.defaults:
+            nodef_args = node.args[:-len(node.defaults)]
+            def_args = node.args[-len(node.defaults):]
+        else:
+            nodef_args = node.args
+            def_args = []
+
+        result = [arg.arg for arg in nodef_args]
+        result.extend('%s=%s' % (arg.arg, default.src) for arg, default in zip(def_args, node.defaults))
+        if node.vararg:
+            result.append('*%s' % node.vararg.arg)
+        if node.kwarg:
+            result.append('**%s' % node.kwarg.arg)
+        return ', '.join(result)
+    def postarg(translator, node):
+        return node.arg
     @priority(14)
     def postOr(translator, node):
         return ' or '.join(expr.src for expr in node.values)
@@ -207,9 +215,9 @@ class PythonTranslator(ASTTranslator):
         return '.'.join((node.value.src, node.attr))
     def postCall(translator, node):
         node.priority = 2
-        args = [ arg.src for arg in node.args ]  + [ kw.src for kw in node.keywords ]
-        if len(args) == 1 and isinstance(node.args[0], ast.GeneratorExp):
-            return node.func.src + args[0]
+        if len(node.args) == 1 and isinstance(node.args[0], ast.GeneratorExp):
+            return node.func.src + node.args[0].src
+        args = [ arg.src for arg in node.args ] + [ kw.src for kw in node.keywords ]
         return '%s(%s)' % (node.func.src, ', '.join(args))
     def postkeyword(translator, node):
         if node.arg is None:
@@ -219,15 +227,17 @@ class PythonTranslator(ASTTranslator):
         return '*' + node.value.src
     def postSubscript(translator, node):
         node.priority = 2
-        if isinstance(node.slice, ast.Tuple):
-            key = ', '.join([elt.src for elt in node.slice.elts])
-        elif isinstance(node.slice, ast.Constant) and isinstance(node.slice.value, tuple):
-            key = repr(node.slice.value)[1:-1]
+        x = node.slice
+        if isinstance(x, ast.Index):
+            x = x.value
+        if isinstance(x, ast.Tuple):
+            key = ', '.join([elt.src for elt in x.elts])
+        elif isinstance(x, ast.Constant) and isinstance(x.value, tuple):
+            key = repr(x.value)[1:-1]
         else:
-            key = node.slice.src
+            key = x.src
         return '%s[%s]' % (node.value.src, key)
-    def postIndex(translator, node):
-        # Python 3.8 and lower
+    def postIndex(translator, node):  # Python <= 3.7
         return node.value.src
     def postSlice(translator, node):
         result = []
@@ -247,11 +257,17 @@ class PythonTranslator(ASTTranslator):
             s = str(value)
             if float(s) == value: return s
         return repr(value)
-    def postNum(translator, node):
+    def postNameConstant(translator, node):  # Python <= 3.7
+        return repr(node.value)
+    def postNum(translator, node):  # Python <= 3.7
         node.priority = 1
         return repr(node.n)
-    def postEllipsis(translator, node):
-        return '...'
+    def postStr(translator, node):  # Python <= 3.7
+        node.priority = 1
+        return repr(node.s)
+    def postBytes(translator, node):  # Python <= 3.7
+        node.priority = 1
+        return repr(node.s)
     def postList(translator, node):
         node.priority = 1
         return '[%s]' % ', '.join(item.src for item in node.elts)
@@ -260,50 +276,34 @@ class PythonTranslator(ASTTranslator):
         if len(node.elts) == 1:
             return '(%s,)' % node.elts[0].src
         return '(%s)' % ', '.join(elt.src for elt in node.elts)
-    def postAssTuple(translator, node):
-        node.priority = 1
-        if len(node.nodes) == 1: return '(%s,)' % node.nodes[0].src
-        else: return '(%s)' % ', '.join(item.src for item in node.nodes)
     def postDict(translator, node):
         node.priority = 1
         return '{%s}' % ', '.join('%s:%s' % (key.src, value.src) for key, value in zip(node.keys, node.values))
     def postSet(translator, node):
         node.priority = 1
         return '{%s}' % ', '.join(item.src for item in node.nodes)
-    def postBackquote(translator, node):
-        node.priority = 1
-        return '`%s`' % node.expr.src
     def postName(translator, node):
         node.priority = 1
         return node.id
-    def preStr(self, node):
-        if self.top_level_f_str is None:
-            self.top_level_f_str = node
-    def postStr(self, node):
-        if self.top_level_f_str is node:
-            self.top_level_f_str = None
-            return "f%r" % ('{%s}' % node.value.src)
-        return '{%s}' % node.value.src
-    def preJoinedStr(self, node):
-        if self.top_level_f_str is None:
-            self.top_level_f_str = node
     def postJoinedStr(self, node):
-        result = ''.join(
-            value.value if isinstance(value, ast.Constant) else '{%s}' % value.src
-            for value in node.values)
-        if self.top_level_f_str is node:
-            self.top_level_f_str = None
-            return "f%r" % result
-        return result
-    def preFormattedValue(self, node):
-        if self.top_level_f_str is None:
-            self.top_level_f_str = node
+        result = []
+        for item in node.values:
+            if isinstance(item, ast.Constant):
+                assert isinstance(item.value, str)
+                result.append(item.value)
+            elif not PY38 and isinstance(item, ast.Str):  # Python 3.7
+                result.append(item.s)
+            elif isinstance(item, ast.FormattedValue):
+                if item.conversion == -1:
+                    src = '{%s}' % item.value.src
+                else:
+                    src = '{%s!%s}' % (item.value.src, chr(item.conversion))
+                result.append(src)
+            else:
+                assert False
+        return "f%r" % ''.join(result)
     def postFormattedValue(self, node):
-        res = '{%s:%s}' % (node.value.src, node.fmt_spec.src)
-        if self.top_level_f_str is node:
-            self.top_level_f_str = None
-            return "f%r" % res
-        return res
+        return node.value.src
 
 
 nonexternalizable_types = (ast.keyword, ast.Starred, ast.Slice, ast.List, ast.Tuple)
@@ -352,7 +352,12 @@ class PreTranslator(ASTTranslator):
     def preLambda(translator, node):
         #if node.varargs or node.kwargs or node.defaults:
         #    throw(NotImplementedError)
-        translator.contexts.append(set(arg.arg for arg in node.args.args))
+        context = set(arg.arg for arg in node.args.args)
+        if node.args.vararg:
+            context.add(node.args.vararg.arg)
+        if node.args.kwarg:
+            context.add(node.args.kwarg.arg)
+        translator.contexts.append(context)
         translator.dispatch(node.body)
         translator.contexts.pop()
         return True
@@ -377,7 +382,11 @@ class PreTranslator(ASTTranslator):
         node.external = True
     def postConstant(translator, node):
         node.external = node.constant = True
-    def postNum(translator, node):
+    def postNum(translator, node):  # Python <= 3.7
+        node.external = node.constant = True
+    def postStr(translator, node):  # Python <= 3.7
+        node.external = node.constant = True
+    def postBytes(translator, node):  # Python <= 3.7
         node.external = node.constant = True
     def postDict(translator, node):
         node.external = True
@@ -385,6 +394,8 @@ class PreTranslator(ASTTranslator):
         node.external = True
     def postkeyword(translator, node):
         node.constant = node.value.constant
+    def postIndex(translator, node):  # Python <= 3.7
+        node.constant = node.value.constant
     def postCall(translator, node):
         func_node = node.func
         if not func_node.external:
@@ -442,7 +453,8 @@ def create_extractors(code_key, tree, globals, locals, special_functions, const_
                 def extractor(globals, locals):
                     return locals['.0']
             else:
-                code = compile(src, src, 'eval')
+                filename = '<pony ' + src + '>'
+                code = compile(src, filename, 'eval')
                 def extractor(globals, locals, code=code):
                     return eval(code, globals, locals)
             extractors[src] = extractor
